;;;;;;;;;;;;;;;;;;;;;;;;;;; -*- Mode: Emacs-Lisp -*- ;;;;;;;;;;;;;;;;;;;;;;;;;;
;; pvs-validate.el -- 
;; Author          : Sam Owre
;; Created On      : Tue Apr  9 11:28:46 1996
;; Last Modified By: Sam Owre
;; Last Modified On: Tue Apr  9 19:33:08 1996
;; Update Count    : 1
;; Status          : Unknown, Use with caution!
;; 
;; HISTORY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defpvs pvs-validate-proofs validation (pvsfile)
  "Validate the proofs of all the formulas in the specified PVS file.

Attempts all formulas in the specified PVS file which have an associated
proof script, and compares the results with earlier runs."
  (interactive (complete-pvs-file-name "Validate proofs for: "))
  (confirm-not-in-checker)
  (popper-bury-output t)
  (save-some-pvs-buffers)
  (let ((vdir (format "%s.validation" *pvs-current-directory*)))
    (unless (file-exists-p vdir)
      (make-directory vdir))
    (let ((vfiles
	   (pvs-send-and-wait (format "(validate-proofs \"%s\")" pvsfile)
			      nil 'vprf 'list)))
      (if (and (car vfiles) (not (eq (car vfiles) 'NIL)))
	  (let ((default-directory *pvs-current-directory*))
	    (apply (function pvs-compare-validation-files) vfiles))
	  (message
	   "No previous proof validation file exists to compare to.")))))

(defun pvs-compare-validation-files (file1 file2)
  (let* ((buf1 (find-file-noselect file1))
	 (buf2 (find-file-noselect file2))
	 (point1 1)
	 (point2 1)
	 (dpoint (abs (compare-buffer-substring
		       buf1 point1 nil buf2 point2 nil))))
    (while (not (zerop dpoint))
      (save-excursion
	(set-buffer buf1)
	(goto-point dpoint)
	
